#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <fcntl.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/wait.h>

#define READ 0
#define WRITE 1
#define ERR 2

#define GRADM_PATH "/sbin/gradm"

static pid_t wrap_process(const char *path, char *const argv[], int *infd,
		   int *outfd, int *errfd)
{
	int p_stdin[2], p_stdout[2], p_stderr[2];
	pid_t pid = -1;
	int ret;

	ret = pipe(p_stdin);
	if (ret)
		return -1;

	ret = pipe(p_stdout);
	if (ret)
		goto out_error;

	ret = pipe(p_stderr);
	if (ret)
		goto out_error2;

	pid = fork();
	if (pid < 0)
		goto out_error3;
	else if (pid == 0) {
		close(p_stdin[WRITE]);
		ret = dup2(p_stdin[READ], READ);
		if (ret < 0)
			goto dup_error;
		close(p_stdout[READ]);
		ret = dup2(p_stdout[WRITE], WRITE);
		if (ret < 0)
			goto dup_error;
		close(p_stderr[READ]);
		ret = dup2(p_stderr[WRITE], ERR);
		if (ret < 0)
			goto dup_error;
		execv(path, argv);
dup_error:
		exit(1);
	}

	if (infd != NULL)
		*infd = p_stdin[WRITE];
	else
		close(p_stdin[WRITE]);
	if (outfd != NULL)
		*outfd = p_stdout[READ];
	else
		close(p_stdout[READ]);
	if (errfd != NULL)
		*errfd = p_stderr[READ];
	else
		close(p_stderr[READ]);

	close(p_stdin[READ]);
	close(p_stdout[WRITE]);
	close(p_stderr[WRITE]);

	return pid;

out_error3:
	close(p_stderr[READ]);
	close(p_stderr[WRITE]);
out_error2:
	close(p_stdout[READ]);
	close(p_stdout[WRITE]);
out_error:
	close(p_stdin[READ]);
	close(p_stdin[WRITE]);
	return pid;
}

// returns -1 on error, 0 on RBAC disabled, 1 on RBAC enabled
int check_rbac_status(void)
{
	int outfd;
	int status;
	pid_t pid;
	char buf[64];
	char *const argv[] = { GRADM_PATH, "-S", NULL };
	int ret = -1;

	pid = wrap_process(GRADM_PATH, argv, NULL, &outfd, NULL);
	if (pid < 0)
		return ret;
	
	ret = 1;

	memset(buf, 0, sizeof(buf));
	read(outfd, buf, sizeof(buf)-1);
	if (strcmp(buf, "The RBAC system is currently enabled.\n"))
		ret = 0;

	close(outfd);
	waitpid(pid, &status, 0);

	return ret;
}

// returns -1 on error, 0 otherwise
int deauthenticate_from_special_role(void)
{
	int status;
	pid_t pid;
	char *const argv[] = { GRADM_PATH, "-u", NULL };

	pid = wrap_process(GRADM_PATH, argv, NULL, NULL, NULL);
	if (pid < 0)
		return -1;
	
	waitpid(pid, &status, 0);

	return 0;
}

// returns -1 on error, 0 on RBAC disabled, 1 on RBAC enabled and failed
// password, 2 on RBAC enabled and successful auth
int authenticate_to_special_role(const char *rolename, const char *password)
{
	int infd, outfd, errfd;
	// handle -a or -n
	char *argv[] = { GRADM_PATH, NULL, (char *)rolename, NULL };
	char buf[64];
	pid_t pid;
	int ret;
	int status;
	int require_auth = password != NULL && strcmp(password, "");

	ret = check_rbac_status();
	// if error or disabled, return that status
	if (ret != 1)
		return ret;

	if (require_auth)
		argv[1] = "-a";
	else
		argv[1] = "-n";

	pid = wrap_process(GRADM_PATH, argv, &infd, &outfd, &errfd);
	if (pid < 0)
		return -1;

	ret = 2;
	if (require_auth) {
		memset(buf, 0, sizeof(buf));
		read(outfd, buf, sizeof(buf)-1);
		if (strcmp(buf, "Password: ")) {
			ret = -1;
			goto out;
		}
		write(infd, password, strlen(password));
		write(infd, "\n", 1);

		memset(buf, 0, sizeof(buf));
		read(errfd, buf, sizeof(buf)-1);
		if (!strcmp(buf, "Invalid password.\n"))
			ret = 1;
	}

out:
	close(infd);
	close(outfd);
	close(errfd);

	waitpid(pid, &status, 0);

	return ret;
}
